
#include <sys/param.h>
#include <sys/kernel.h>
#include <sys/systm.h>
#include <sys/timebase.h>

static struct timebase* timebase_system;

u_int64_t timebase_count;
int64_t timebase_us_per_count;
int64_t timebase_delta_us;

void
timebase_register(struct timebase *tb)
{
	int s = splclock();

	timebase_system = tb;
	timebase_count = 0;

	splx(s);
}

int
timebase_microtime(struct timeval *tv)
{
	u_int64_t count;
	int64_t tmp;
	int delta_us = 0;
	int64_t delta_count;

	if(timebase_system == 0 || timebase_count == 0)
		return (0);

	*tv = time;
	
	count = timebase_system->tb_read(timebase_system);

	delta_count = count - timebase_count;

	/* Convert from counts to to fixed-point microseconds */
	tmp = delta_count * timebase_us_per_count;

	tmp += timebase_delta_us;	/* include previous error */

	/*
	 * Extract the integer us and round to nearest
	 */
 	delta_us = (tmp + (1ll << 29)) >> 30;

	tv->tv_usec += delta_us;

	while(tv->tv_usec > 1000000) {
		tv->tv_sec++;
		tv->tv_usec -= 1000000;
	}

	return (1);
}

int
timebase_delta()
{
	u_int64_t count;
	int64_t tmp;
	int delta_us = 0;
	int64_t delta_count;
	extern int ticks;

	if(timebase_system == 0)
		return (tick);

	count = timebase_system->tb_update(timebase_system);

	if(timebase_count == 0) {
		timebase_count = count;
		timebase_delta_us = 0;
		timebase_us_per_count = (1000000ll << 30) / timebase_system->tb_frequency; 
		return (tick);
	}

	delta_count = count - timebase_count;

	if(delta_count < 0) {
		printf("\n***** TIMEBASE OVERFLOW ****\n");
		delta_count = 0;
	}

	tmp = delta_count * timebase_us_per_count;

	tmp += timebase_delta_us;	/* include previous error */

	/*
	 * Extract the integer microseconds and round to nearest
	 */
 	delta_us = (tmp + (1ll << 29)) >> 30;

	tmp -= ((int64_t)delta_us) << 30;

	if(tmp > (1 << 30) || tmp < -(1 << 30))
	{
		static int i;
		if(++i < 8)
			printf("tmp out of range: %llx\n", tmp);
	}

	/*
	 * Save the accumulated error
	 */
	timebase_delta_us = tmp;

        if(delta_us > (tick + tick / 4) || delta_us < (tick - tick / 4)) {
                static int i;
                static int last;

                if(++i < 8)
                        printf("crazy delta (%d): %d vs. %d (0x%16.16llx counts at 0x%16.16llx->0x%16.16llx)\n",
			    i, delta_us, tick, delta_count, timebase_count, count);

                if(last) {
                        if(last + hz * 5 < ticks)
                                i = 0;
                }
                last = ticks;
        }

	timebase_count = count;
	return delta_us;
}

